/**********************************************************************
 **                                                                  **
 **                               ebug.h                             **
 **                                                                  **
 **  Header file for ebug.                                           **
 **                                                                  **
 **  Written by:  Yoichiro Endo (Mobile Robot Lab, Georgia Tech)     **
 **                                                                  **
 **  Copyright 2000 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: ebug.h,v 1.2 2000/10/25 05:43:02 endo Exp endo $ */

#ifndef EBUG_H
#define EBUG_H

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

#include <Xm/Xm.h>
#include <Xm/MwmUtil.h>
#include <Xm/ArrowBG.h>
#include <Xm/BulletinB.h>
#include <Xm/CascadeB.h>   
#include <Xm/DialogS.h>
#include <Xm/DrawingA.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/MainW.h>
#include <Xm/MessageB.h>
#include <Xm/PanedW.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/Scale.h>
#include <Xm/ScrollBar.h>
#include <Xm/ScrolledW.h>
#include <Xm/Separator.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/ToggleB.h>

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <malloc.h>
#include <ctype.h>
#include <string.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <getopt.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>

#define BITMAPDEPTH        1
#define TOO_SMALL          0
#define BIG_ENOUGH         1

#define DELAY              1000000
#define DEFAULT_X          100
#define DEFAULT_Y          100
#define DEFAULT_THETA      0
#define DEFAULT_TRAILX     0
#define DEFAULT_TRAILY     80
#define MAX_SEGMENTS       100
#define MAX_LIGHTS         1000
#define UPDATE_CYCLE       0.01

#define NT_POSITIVE        1
#define NT_NEGATIVE        2
#define NT_THRESHHOLD      3
#define NT_THRESHHOLDNEG   4

#define FALSE               0
#define TRUE                1
#define ERASE               0
#define DRAW                1
#define LEFT                0
#define RIGHT               1
#define FRONT               2
#define REAR                3
#define BRAITENBERG_VEHICLE 0
#define TOLMAN_SOWBUG       1
#define PRIMARY             1
#define SECONDARY           2
#define ENCODE              100
#define DECODE              0.01
#define SIM_TO_REAL         0.01 
#define POSITIVE            1
#define NEUTRAL             0
#define NEGATIVE           -1
#define TOGGLE              2
#define ORIENTATION         0
#define PROGRESSION         1
#define HEADING             2
#define PARTIAL             0
#define ALL                 1

#define BIGNUMBER          (1.0e7)
#define SMALLNUMBER        (1.0e-7)

#define NO_TYPE           -1
#define TYPE_A             0
#define TYPE_B             1
#define NUM_TYPES          2

#define NUM_VECTORS        2

#define HYPOTHESIS_MIN                  0
#define HYPOTHESIS_MAX                  100
#define HYPOTHESIS_DEFAULT_VALUE        50
#define HYPOTHESIS_INC                  10
#define ORIENTATION_NEED_MIN            0
#define ORIENTATION_NEED_MAX            100
#define ORIENTATION_NEED_DEFAULT_VALUE  50
#define ORIENTATION_NEED_INC            5

#define MAX_WORKPROCS      4
#define XDRAWARCUNITS      64 // 64 = 1 degree

#define SIMULATION         0
#define EBUG               0
#define HSERVER            1
#define LINEAR             0
#define ANGULAR            1

extern Display            *display;
extern Widget             top_level;
extern Widget             main_window, scrolled_window;
extern Widget             work_area, drawing_area;
extern Widget             command_bar, menu_bar;
extern Widget             load_lights_button, run_pause_button, reset_button, quit_button;
extern Widget             hypothesis_window_shell, hypothesis_window, hypothesis_scale[NUM_TYPES];
extern Widget             orientation_need_window_shell, orientation_need_window, orientation_need_scale[NUM_TYPES];
extern Pixmap             drawing_area_pixmap, plane;
extern Colormap           ebug_colormap;
extern GC                 gc;
extern GC                 blueline1gc, blueline3gc;
extern GC                 lightblueline1gc, lightblueline3gc;
extern GC                 redline1gc, redline3gc;
extern GC                 pinkline1gc, pinkline3gc;
extern GC                 trailgc;
extern GC                 sensorgc;
extern GC                 lightgc[NUM_TYPES];
extern GC                 vectorgc[NUM_VECTORS][NUM_TYPES];
extern GC                 graphgc[NUM_VECTORS][NUM_TYPES];
extern GC                 theGC;
extern GC                 xorGC;
extern GC                 red_xorGC;
extern Pixel              red_pixel, pink_pixel, blue_pixel, lightblue_pixel, green_pixel;
extern Pixel              lightcolor_pixel[NUM_TYPES];
extern Pixel              vector_pixel[NUM_VECTORS][NUM_TYPES];
extern Pixel              graph_pixel[NUM_VECTORS][NUM_TYPES];
extern int                screen_num;
extern XtWorkProc         workproc[MAX_WORKPROCS];
extern XtAppContext       app;
extern XEvent             event;

extern struct t_node      *robotnameList, *robotList, *lightList;
extern unsigned int       width, height;

extern double             birth_time;
extern double             age, delta_time; 

extern int                run_type;
extern unsigned long      event_mask;
extern bool               show_trails;
extern bool               load_default_robot;
extern bool               load_default_light;
extern bool               eat_lights;
extern bool               debug;
extern bool               log_data;
extern char               *default_robot;
extern char               *default_light;

extern int                go;
extern int                trial;
extern int                lightId;

extern char               *log_filename;
extern FILE               *logfile;

extern struct IPCommunicator *communicator;
extern struct IPConnection   *console;

//-----------------------------------------------------------------------------
// Structures

struct point {
  int x;
  int y;
};

struct t_node {
  struct t_node *next;
  int *data;
};

struct t_light {
  int id;
  int x,y;
  int lum;
  int radius;
  int status;
  int type;
  double age;
};

struct t_motor {
  int id;
  int x,y;
  double angle;
  double strength;
  double force;
  double maxForce;
};

struct t_sensor {
  int id;
  int x,y;
  double angle;
  double fov;
  double sensitivity;
  double stimulus;  
};

struct t_vector {
  int id;
  int disp_x, disp_y;
  int side;
  int type;
  double value;
};

struct t_photo_sensor {
  int id;
  int sensor_x, sensor_y;
  double angle;
  double fov;
  double sensitivity;
  int side;
  double value[NUM_TYPES];
};

struct t_line {
  int x1,y1;
  int x2,y2;
};

struct t_neuron {
  int sensorId;
  int motorId;
  int neuronType;
};

struct t_robot {
  char name[100];
  int id;
  int type;
  double x,y;
  double last_x, last_y;
  int width;
  double mass;
  double theta;
  double last_theta;
  double trailx, traily;
  double angvel,speed;
  double last_angvel;
  double progression;
  int vte;
  int cmx,cmy;
  unsigned int bug_height, bug_width;
  double maxTorque;
  int last_seen_light_type;
  double hypothesis[NUM_TYPES];
  double orientation_need[NUM_TYPES];
  struct t_node *lineList;
  struct t_node *motorList;
  struct t_node *sensorList;
  struct t_node *vectorList[NUM_VECTORS][NUM_TYPES];
  struct t_node *photo_sensorList;
  struct t_node *neuronList;
  struct t_node *trailList;
};

//-----------------------------------------------------------------------------
// Function calls

// Initializing the experiment
void initializeExperiment(int init_type);

// Main Loop
void ebugMainLoop(void);

// Callback functions
void kb_input_cb(Widget w, XtPointer data, XmDrawingAreaCallbackStruct *callback_data);
void updateHypothesisValue_cb(Widget w, XtPointer client_data, XtPointer call_data);
void updateOrientationNeedValue_cb(Widget w, XtPointer client_data, XtPointer call_data);
void pushLoadLightsButton_cb(Widget w, XtPointer client_data, XtPointer call_data);
void pushRunPauseButton_cb(Widget w, XtPointer client_data, XtPointer call_data);
void pushResetButton_cb(Widget w, XtPointer client_data, XtPointer call_data);
void pushQuitButton_cb(Widget w, XtPointer client_data, XtPointer call_data);

//-----------------------------------------------------------------------------
// classes

class ScreenUtility{

public:
  ScreenUtility();
  Widget MainWindow(Widget parent);
  Widget MenuBar(Widget parent);
  Widget WorkArea(Widget parent);
  Widget CommandBar(Widget parent);
  Widget LoadLightsButton(Widget parent);
  Widget RunPauseButton(Widget parent, Widget left_widget);
  Widget ResetButton(Widget parent, Widget left_widget);
  Widget QuitButton(Widget parent, Widget left_widget);
  Widget ScrolledWindow(Widget parent, Widget top_widget);
  Widget DrawingArea(Widget parent, unsigned int width, unsigned int height);
  Widget HypothesisWindow(Widget parent);
  Widget OrientationNeedWindow(Widget parent);
  void createWindows(unsigned int width, unsigned int height);
  void clearWindow(void);
  void redrawWindow(void);
  void refreshWindow(void);
  void initializeX(void);
  void initializeDrawingArea(void);
  void updateRunPauseButton(void);
  void updateHypothesisWindow(void);
  void updateOrientationNeedWindow(void);
  void getXEvent(void);
  void getGC(Window win, XFontStruct *font_info);
  void saveLights(t_node *lightList);
  void load_font(XFontStruct **font_info);
  void drawRobots(int draw, int window_num);
  void drawLights(int draw);
  void drawWindow2box(Drawable win, GC gc);
  void drawSensors(t_robot *robot, int draw, int window_num);
  void drawVectors(t_robot *robot, int vector_type, int draw, int window_num);
  void drawInternalGraph(t_robot *robot, int vector_type, int draw, int window_num);
  void drawTrail(t_robot *robot, int draw);
  void constructArrowHead(XSegment *segments, int side, int *i);
  void translateSegments(XSegment segments[], int nsegments, double x, double y);
  void convertSegments(XSegment segments[], int nsegments);
  void rotateSegments(double theta, XSegment segments[], int nsegments, double cx, double cy);
  void displayAge(void);
  void displayTrials(void);
  void displayVTE(void);
  void drawText(GC gc, char *text, int x, int y, int draw);
  Pixmap createColoredPixmap(int width, int height, unsigned char *bits, Pixel fg, Pixel bg);
  void buildBitMaps(Pixel fg, Pixel bg);
};

//-----------------------------------------------------------------------------
class HardwareCommand{

public:
  HardwareCommand();
  void initialize(int init_type);
  void getCurrentRobotInfo(void);
  void getCurrentLightInfo(void);
  void driveRobot(void);
  void stopRobot(void);
  void disconnectRobot(void);
};
//-----------------------------------------------------------------------------
class Calculate{

public:
  Calculate();
  void convertTheta(double *theta);
  void convertCoordinate(double *x, double *y, double *theta, int in, int out);
  void convertVelocity(double *vel, int type, int in, int out);
  double PhotoSensorValue(struct t_robot *robot, struct t_photo_sensor *photo_sensor, int light_type);
  void Vectors(int vector_type);
  double OrientationVector(struct t_robot *robot, int side, int light_type);
  double ProgressionVector(struct t_robot *robot, int side, int light_type);
  void Hypothesis(void);
  void OrientationNeed(void);
  void AngularSpeed(struct t_robot *robot);
  void ForwardSpeed(struct t_robot *robot);
  void VTE(void);
  int signPN(double value);
  double diffAngle(double ang1, double ang2);
  int inFieldOfView(double smin, double smax, double ang);
  void rotatePoint(double theta, double *x, double *y, double cx, double cy);

  // For Braitenberg Vehicles
  void BVsensorValues(void);
  void BVmotorForce(struct t_robot *robot);
  void BVclearMotorForce(struct t_robot *robot);
  void BVtotalForce(struct t_robot *robot, double *force, double *angle);
  void BVtorque(struct t_robot *robot, double *angvel);
};

//-----------------------------------------------------------------------------
class ListUtility{

public:
  ListUtility();
  void push(struct t_node **list, int *data);
  void clearList(struct t_node **list);
  void create(struct t_node **list);
};
//-----------------------------------------------------------------------------
class TimeUtility{

public:
  TimeUtility();
  char *currentDate(void);
  double currentTime(void);
  void computeAge(void);
  void nap(double nap_time);
};
//-----------------------------------------------------------------------------
class RobotFunction{

public:
  RobotFunction();
  void initialize(int init_type);
  void moveRobot(struct t_robot *robot);
  void driveRobot(void);
  int isMoving(void);
  void load(struct t_robot *robot, char *filename);

  // For Braitenberg Vehicles
  void BVmoveRobot(double *x, double *y, double theta, double distance);
};

//-----------------------------------------------------------------------------
class LightFunction{

public:
  LightFunction();
  void initialize(void);
  int inLight(point pt, int *id);
  t_light *closestLight(t_robot *robot, double *closest_dist);
  int lookingLightDirectly(t_robot *robot, int light_type);
  void changeStatus(int id, int status);
  void changeType(int id);
  int lightType(int id);
  void changeSize(int id);
  void updateLightAge(void);
  void deleteOldLights(double max_old);
  void remove(int id);
  void add(point pt);
  void load(char *filename);
  void save(void);
};

//-----------------------------------------------------------------------------
class LogUtility{

public:
  LogUtility();
  void initializeLogFile(void);
  void recordPerformance(void);
  void closeLogFile(void);
};

#endif
/**********************************************************************
 * $Log: ebug.h,v $
 * Revision 1.2  2000/10/25 05:43:02  endo
 * Minor change.
 *
 * Revision 1.1  2000/10/25 04:45:27  endo
 * Initial revision
 *
 * Revision 1.30  2000/08/13 09:01:31  endo
 * getCurrentLightInfo added.
 *
 * Revision 1.29  2000/08/10 06:15:39  endo
 * driveRobot, stopRobot etc.
 *
 * Revision 1.28  2000/08/08 09:09:41  endo
 * HServer implemented.
 *
 * Revision 1.27  2000/08/07 21:01:38  endo
 * Reset feature added.
 *
 * Revision 1.26  2000/08/07 01:42:03  endo
 * Orientation Need Window created.
 *
 * Revision 1.25  2000/08/07 00:56:13  endo
 * ScreenUtility::simulateRobots changed as simbugMainLoop.
 *
 * Revision 1.24  2000/08/06 07:26:41  endo
 * Two light types implemented successfully.
 *
 * Revision 1.23  2000/07/31 05:31:17  endo
 * Hypothesis feature improved. Orientation Need feature added.
 *
 * Revision 1.22  2000/07/29 22:17:58  endo
 * Reward stuff implemented.
 *
 * Revision 1.21  2000/05/02 13:38:45  endo
 * Minor cleaning up.
 *
 * Revision 1.20  2000/04/30 12:42:06  endo
 * Internal Graph Implemented. Not perfect, though.
 *
 * Revision 1.19  2000/04/30 08:03:36  endo
 * photo_sensor implemented.
 *
 * Revision 1.18  2000/04/30 01:35:15  endo
 * ArrowHeads success.
 *
 * Revision 1.17  2000/04/29 23:13:32  endo
 * DrawTrail worked.
 *
 * Revision 1.16  2000/04/29 09:18:19  endo
 * Progression Vector working.
 *
 * Revision 1.15  2000/04/29 07:31:53  endo
 * Hypothesis Window added.
 *
 * Revision 1.14  2000/04/28 04:38:53  endo
 * secondary robot implemented.
 *
 * Revision 1.13  2000/04/27 21:27:09  endo
 * buttons added.
 *
 * Revision 1.12  2000/03/27 00:48:55  endo
 * Trying to implement drawVectors.
 *
 * Revision 1.11  2000/03/26 10:14:31  endo
 * drawSensors added.
 *
 * Revision 1.10  2000/03/26 09:59:02  endo
 * *** empty log message ***
 *
 * Revision 1.9  2000/03/26 09:57:36  endo
 * Placing lights from mouse clicks were also
 * successful.
 *
 * Revision 1.8  2000/03/26 08:35:18  endo
 * Finally, the simulation works as it supposed to be.
 *
 * Revision 1.7  2000/03/26 05:17:51  endo
 * Got rid of Start. Run/Pause instead.
 *
 * Revision 1.6  2000/03/13 06:37:37  endo
 * Start and Pause works. Getting right input (e.g., key
 * board char, mouse etc.) is still having a problem.
 *
 * Revision 1.5  2000/03/12 09:12:37  endo
 * got rid of pthread, since it doesn't do any
 * good.
 *
 * Revision 1.4  2000/03/12 01:36:31  endo
 * kb_input moved to screen.cc
 *
 * Revision 1.3  2000/03/05 05:59:17  endo
 * pthread implemented.
 *
 * Revision 1.2  2000/03/03 08:28:13  endo
 * Lots of widgets (Motif) added.
 *
 * Revision 1.1  2000/03/03 01:29:52  endo
 * Initial revision
 *
 **********************************************************************/
